(* This program finds the roots of a polynomial/function whose roots are \
inside
the unit circle, by:
 1 Pick a large number of guesses inside the unit circle
 2 Using Newton's method to find the roots near those inital guesses
 3 The previous step may find duplicated roots. This step throws away 
 most the repeated roots, which are at distance less than delta1 away.
 4 Repeat step 2, using the list from step 3 as initial guesses
 5 Repeat step 3 with a smaller distance limit: delta2.
 6 Refine the approximated root list by throwing away numbers, when plug
 in to the polynomial with which, return results with absolute values 
 larger than a predefined small epsilon 
 
 The program finds almost all the roots of Z'(U,z) at high \
precision. 
 (For matrices of size 300, the error returned when plugging in found roots \
is less than 10^-6)
 
 To call function FindZprimes:
 FindZprimes[zlist,delta1,delta2,epsilon];
 Where zlist is the list of U's eigenvalues; delta1 is the argument for step \
3;
 delta2 is the argument for the step 5; epsilon is the argument for step 6. 
 Recommended: delta1=10^-6; delta2=10^-5; epsilon=10^-5;
 For an example, see file example300.nb
 *)

PickInitials[r_, num_] := Block[{counter, theta1},
      theta1 = Random[Real, {0, 2 *Pi/num}];
      initiallist = {r*E^(I*theta1)};
      For[counter = 2, counter <= num, counter++,
        AppendTo[initiallist, r*E^(I*(theta1 + 2*(counter - 1)* Pi/num))];
        ];
      
      initiallist
      ];


FindRawRoots[f_, initguesses_] := Block[{rawlist, i},
      rawlist = {};
      For[i = 1, i <= Length[initguesses], i++,
        rawlist = Append[rawlist, FindRoot[f, {x, initguesses[[i]]}]];
         ];
      x /. rawlist
      ];


TossRepeats[Oldlist_, eps_] := 
    Block[{NoRepeatList, n, j, thiselt, nextelt},
      SortedOldlist = Sort[Oldlist,Abs[#1]<Abs[#2]&];
      n = Length[Oldlist];
      
      For[j=n, j>=1, j=j-1,
        If[Abs[SortedOldlist[[j]]] < 1, 
          SortedOldlistLess1 = Take[SortedOldlist,j];
          Break[];
        ];
      ];
        
      NoRepeatList={SortedOldlistLess1[[1]]};
      n = Length[SortedOldlistLess1];
   
      For[j = 1, j <= n - 1, j = j + 1, thiselt = SortedOldlistLess1[[j]]; 
        nextelt = SortedOldlistLess1[[j + 1]]; 
        If[Abs[nextelt - thiselt] >= eps, 
          NoRepeatList = Append[NoRepeatList, nextelt];
          ];
        ];
      NoRepeatList
      ];


RefineRootList[f_, norepeatlist_, esp_] := Block[{i},
      refinedlist = {};
      For[i = 1, i <= Length[norepeatlist], i++,
        If[
            Abs[f /. x -> norepeatlist[[i]]] <= esp && 
              Abs[norepeatlist[[i]]] < 1,
            refinedlist = Append[refinedlist, norepeatlist[[i]]];
            ];
        ];
      
      refinedlist
      ];


MidPointList[inlist_] := Block[{i, len},
      Sort[inlist];
      midpointlist = {};
      len = Length[inlist];
      For[i = 1, i <= len - 1, i++,
        midpointlist = 
            Append[midpointlist, (inlist[[i]] + inlist[[i + 1]])/2];
        ];
      midpointlist = Append[midpointlist, (inlist[[len]] + inlist[[1]])/2];
      
      midpointlist
      ];


FindZprimes[zlist_, delroot1_, delroot2_, eps_] := Block[{finalRoots, i, n},
      Clear[f];
      n = Length[zlist];
      f = Sum[1/(x - zlist[[i]]), {i, 1, n}];
      
      initguesses = MidPointList[zlist];
      
      For[i = 0, i <= 3, i = i + 0.1,
        initguesses = Union[initguesses, PickInitials[1 - i/n, 2*n]];
        ];
      
      For[i = 3, i <= 4, i = i + 0.1,
        initguesses = Union[initguesses, PickInitials[1 - i/n, n]];
        ];
      
      For[i = 4, i <= 10, i = i + 0.5,
        initguesses = Union[initguesses, PickInitials[1 - i/n, n]];
        ];
      
      For[i = 10, i <= 20, i = i + 2,
        initguesses = Union[initguesses, PickInitials[1 - i/n, Floor[n/2]]];
        ];
      
      For[i = 20, i <= n, i = i + 10,
        initguesses = Union[initguesses, PickInitials[1 - i/n, Floor[n/10]]];
        ];
      
      rawroots = FindRawRoots[f, initguesses];
          
      lessRepeatRoots = TossRepeats[rawroots, delroot1];
     
      lessrawroots=FindRawRoots[f,lessRepeatRoots];
      
      noRepeatRoots = TossRepeats[lessrawroots, delroot2];
      
      finalRoots = RefineRootList[f, noRepeatRoots, eps];
     
      finalRoots
      ];

